import numpy as np


class Labels:
    """
        # TODO: add docstring
    """

    def __init__(self, y):
        """

        Args:
            y: iterable

        """

        self.labels = np.array(y)

        # if one-hot, transform to labels
        if self.labels.ndim > 1:
            self.labels = self.labels.argmax(axis=1)

        # test that there are at least two classes
        assert len(set(self.labels.ravel())) >= 2,\
            "Labels should have at least two distinct observations, got {}".format(set(self.labels.ravel()))



import numpy as np
import pandas as pd


def reshape(array):
    """

    Args:
        array: np.ndarray
            N-dimensional array

    Returns: np.ndarray
        2D array if at least 2 elements

    """
    assert array.ndim in [1, 2],\
        "Expected 1D or 2D array, got {}D".format(array.ndim)

    # If array is 1D, reshape it to be 2D. If 2D, do nothing
    ret = array.reshape(-1, 1) if array.ndim == 1 else array

    # Check that we have at least two lines
    assert ret.shape[0] and ret.shape[1],\
        "Expected at least 2 data points, got {}".format(array.shape[0])

    return ret


def array_from_ndarray(array):
    """

    Args:
        array: np.ndarray

    Returns:

    """
    assert array.dtype in [float, int],\
        "Expected array of int or float, got {}".format(array.dtype)
    return reshape(array)


def preprocess(X):
    """
    Args:
        X: iterable
            Array containing the features

    Return:
        np.ndarray: 2D array
    """

    if isinstance(X, pd.DataFrame):
        res = array_from_ndarray(X.values)
    elif isinstance(X, np.ndarray):
        res = array_from_ndarray(X)
    elif isinstance(X, (list, tuple)):
        res = array_from_ndarray(np.array(X))
    else:
        raise TypeError("Expected an np.array, a pd.Dataframe, or a list.\
                                Got: {}".format(type(X)))

    return res



import unittest

import numpy as np
import pandas as pd

from pycvi.utils import labels

##########################################
#             Test cases                 #
##########################################

VALID = {
    0: (pd.DataFrame([[1, 0, 0], [0, 1, 0], [0, 0, 1]]),
        np.array([0, 1, 2])),
    1: (np.array([[1, 0], [0, 1], [0, 1], [0, 1]]),
        np.array([0, 1, 1, 1])),
    2: ([[0, 1], [1, 0]],
        np.array([1, 0])),
    3: (((0, 1), (1, 0)),
        np.array([1, 0])),
    4: ([1, 2, 3, 4],
        np.array([1, 2, 3, 4]))
}

INVALID = {
    0: ([[], []], ValueError),  # trying to get argmax on an empty sequence
    1: ({1, 0, 2}, TypeError),
    2: ([[1, 0, 0], [1, 0, 0], [1, 0, 0]], AssertionError),  # no two distinct labels
    3: ([1, 1, 1, 1], AssertionError)
}


class TestLabels(unittest.TestCase):
    def test_valid(self):
        for key, value in sorted(VALID.items(), key=lambda x: x[0]):
            self.assertEqual(list(labels.Labels(value[0]).labels.ravel()),
                             list(value[1].ravel()),
                             "{} != {}".format(labels.Labels(value[0]).labels, value[1]))

    def test_invalid(self):
        for key, value in sorted(INVALID.items(), key=lambda x: x[0]):
            self.assertRaises(value[1], labels.Labels, value[0])



import logging
import unittest


from pycvi.external.purity import Purity


y_true0 = [0, 0, 0, 0, 0, 1, 1, 1, 1, 1]

y_pred0 = [0, 0, 0, 0, 0, 1, 1, 1, 1, 1]
y_pred1 = [0, 0, 0, 1, 1, 1, 1, 1, 0, 0]


class TestPurity(unittest.TestCase):
    def test_0(self):
        print("Starting test_0")
        p = Purity(y_true0, y_pred0).purity
        self.assertEqual(1, p)
        p = Purity(y_true0, y_pred1).purity
        self.assertEqual(.6, p)



from . import parse


class Features:
    def __init__(self, X):
        """
        Args:
            X: iterable
                Array containing the features
        """

        self.X = parse.preprocess(X)



from .labels import Labels
from .features import Features


class DataSet(Features, Labels):
    def __init__(self, X, y=None):
        if y is not None:
            Labels.__init__(self, y)
        else:
            self.labels = None
        Features.__init__(self, X)

        assert self.labels.shape[0] == self.X.shape[0],\
            "Expected features and targets to have the same number of elements, got {} and {}".format(
                self.X.shape[0], self.labels.shape[0]
            )



import logging
import numpy as np

from .parse import preprocess

logger = logging.getLogger(__name__)


def mean_mahalanobis(data, mu, sigma):
    """
    Args:
        data: ndarray
            T_i * D array contaning the T_i D-dimensional data points in the cluster
        mu: ndarray
            D-dimensional array representing the mean of the component
        sigma: ndarray
            D*D array representing the covariance matrix of the component.

    Returns:
        S: float
            The mean distance between data points and mu, with covariance matrix sigma.
    """

    logger.info("Computing mean mahalanobis distance")
    print()
    print("Computing mean mahalanobis distance\n")
    print("received data:{}\n".format(data))

    if data.ndim == 1:
        data = data.reshape(1, -1)

    assert sigma.size == data.shape[1] ** 2, \
        "Wrong number of values for sigma, should have {}, got {}".format(data.shape[1] ** 2, sigma.size)

    sigma = sigma.reshape(data.shape[1], -1)

    # print("data: {}".format(data))
    x_minus_mu = data - mu
    inv_sigma = np.linalg.pinv(sigma)
    # print("x_minus_mu.shape: {}".format(x_minus_mu.shape))
    # print("inv_sigma.shape: {}".format(inv_sigma.shape))

    left_term = np.matmul(x_minus_mu, inv_sigma)
    # print("left_term.shape: {}".format(left_term.shape))
    tmp = np.matmul(left_term, x_minus_mu.T).diagonal()
    tmp = tmp ** (1/2)

    # print("tmp: {}".format(tmp))

    return tmp.mean()



import unittest

import numpy as np
import pandas as pd

from pycvi.utils import parse


def arr(value, t='arr'):
    """

    Args:
        value: list
            list to use in order to create the object
        t: str
            'arr' to create an np.array and 'df' to create a pd.DataFrame
    Returns: the array of the appropriate type

    """

    if t == 'df':
        return pd.DataFrame(value)
    return np.array(value)


df_empty1d, df_empty2d, df_empty3d = arr([], t='df'), arr([[], []], t='df'),\
                                     arr([[[], []], [[], []], [[], []]], t='df')
df_scalar = arr([42], 'df')
df_1d = arr([0, 1, 2], 'df')
df_2d = arr([[0, 1], [2, 3]], 'df')
df_3d = arr([[[ 0,  1,  2], [ 3,  4,  5]],
             [[ 6,  7,  8], [ 9, 10, 11]],
             [[12, 13, 14], [15, 16, 17]]], 'df')

arr_empty1d, arr_empty2d, arr_empty3d = [np.zeros((0,)*i) for i in range(1, 4)]
arr_scalar, arr_1d, arr_2d, arr_3d = \
    arr([42]), arr([0, 1, 2]), arr([[0, 1], [2, 3]]),\
    arr([[[ 0,  1,  2], [ 3,  4,  5]],
         [[ 6,  7,  8], [ 9, 10, 11]],
         [[12, 13, 14], [15, 16, 17]]])

one_data_point_2d = [
    arr([[42, 52]], 'df'),
    arr([[42, 52]]),
    [[42, 52]]
]


empty_arrays = [
                df_empty1d, df_empty2d, df_empty3d,
                arr_empty1d, arr_empty2d, arr_empty3d,
                [], [[], []], [[[], []], [[], []]]
               ]

scalar_arrays = [df_scalar, arr_scalar, [42]]

one_d_arrays = [
                df_1d, arr_1d, [0, 1, 2]
               ]

two_d_arrays = [
                df_2d, arr_2d, [[0, 1], [2, 3]]
               ]

three_d_arrays = [
    df_3d, arr_3d, [[[ 0,  1,  2], [ 3,  4,  5]],
                    [[ 6,  7,  8], [ 9, 10, 11]],
                    [[12, 13, 14], [15, 16, 17]]]
]

incorrect_type = [
    42,
    'Bonjour',
    50.7,
    {1, 2, 3},
    {'a': 0, 'b': 1}
]


class TestPreprocess(unittest.TestCase):
    def test_empty_arrays(self):
        for e in empty_arrays:
            self.assertRaises(AssertionError, parse.preprocess, e)

    def test_1d_arrays(self):
        for e in one_d_arrays:
            self.assertSequenceEqual(list(parse.preprocess(e)),
                                     list(np.array([[0], [1], [2]])),
                                     "fail test 1d")

    def test_2d_arrays(self):
        for e in two_d_arrays:
            self.assertSequenceEqual(list(parse.preprocess(e).ravel()),
                                     list(np.array([[0, 1], [2, 3]]).ravel()),
                                     "fail test 2d")

    def test_3d_arrays(self):
        for e in three_d_arrays:
            self.assertRaises(AssertionError, parse.preprocess, e)

    def test_incorrect_type(self):
        for e in incorrect_type:
            self.assertRaises(TypeError, parse.preprocess, e)



from pycvi.internal import davies_bouldin

X0 = [1, 2, 3, 4, 5]
X1 = [[1, 2], [3, 4], [5, 6], [7, 8], [9, 10]]
y0 = [0, 0, 0, 1, 1]
y1 = [1, 0, 0, 1, 1]

a = davies_bouldin.DB(X0, y0)
print(a.db)

b = davies_bouldin.DB(X0, y1)
print(b.db)

c = davies_bouldin.DB(X1, y0)
print(c.db)


import logging
import pathlib

import numpy as np

from itertools import product
from ..utils import dataset, distances


filename = str(pathlib.Path(__file__ + '/../../tmp.log').resolve())

print("log file: {}".format(filename))
logging.basicConfig(filename=filename,
                    filemode='w',
                    format="%(levelname)s:%(asctime)s:%(name)s: %(message)s \n",
                    level=logging.DEBUG)
logger = logging.getLogger(__name__)


class DB(dataset.DataSet):
    def __init__(self, X, y):
        logger.info("creating DB instance")
        super().__init__(X, y)
        logger.debug("self.X.shape = {} and self.labels.shape = {}".format(self.X.shape, self.labels.shape))

        self.unique_labels = sorted(np.unique(self.labels))
        logger.debug("self.unique_labels = {}".format(self.unique_labels))

        self.K = len(self.unique_labels)
        logger.debug("self.K = {}".format(self.K))

        self.kmu, self.ksigma = self.compute_kmu_ksigma()
        logger.debug("self.kmu = {} \nself.ksigma = {}".format(self.kmu, self.ksigma))

        self.db = None

        self.fit()

    def fit(self):
        logging.info("Starting computation of DB index")

        S = [
            distances.mean_mahalanobis(self.X[self.labels == k],
                                       self.kmu[k],
                                       self.ksigma[k])
            for k in self.unique_labels
        ]
        logging.debug("S = {}".format(S))

        M = np.zeros((self.K, self.K))
        for i, j in product(range(self.K), range(self.K)):
            print("Calling mean mahalanobis with data:\n{}\n".format(self.kmu[j]))
            M[i, j] = (i != j) and distances.mean_mahalanobis(self.kmu[j], self.kmu[i], self.ksigma[i])
        logging.debug("M = {}".format(M))

        R = np.zeros((self.K, self.K))
        for i, j in product(range(self.K), range(self.K)):
            R[i, j] = (i != j) and (S[i] + S[j]) / M[i, j]

        self.db = (1/self.K) * R.max(axis=1, initial=0).sum()

    def compute_kmu_ksigma(self):
        mu = [np.mean(self.X[self.labels == k], axis=0) for k in self.unique_labels]
        sigma = [np.cov(self.X[self.labels == k], rowvar=False) for k in self.unique_labels]

        return mu, sigma



import numpy as np

from sklearn import metrics
from ..utils.labels import Labels


class Purity:
    """"""

    def __init__(self, y_true, y_pred):
        self.y_true, self.y_pred = Labels(y_true), Labels(y_pred)
        self.purity = None
        self.fit()

    def fit(self):
        """
        compute the purity score of a clustering

        Returns: float
                purity score.

        """

        # if no value was past to clusters, use the last clustering

        # compute contingency matrix (also called confusion matrix)
        contingency_matrix = metrics.cluster.contingency_matrix(self.y_true.labels, self.y_pred.labels)

        self.purity = np.sum(np.amax(contingency_matrix, axis=0)) / np.sum(contingency_matrix)



import unittest

import numpy as np
import pandas as pd

from pycvi.utils import dataset

##########################################
#             Test cases                 #
##########################################

VALID = {
    0: {
        "X": pd.DataFrame([[0, 1], [1, 2]]),
        "y": [1, 2],
        "XX": np.array([[0, 1], [1, 2]]),
        "yy": np.array([1, 2])
    }
}

INVALID = {
    0: {
        "X": pd.DataFrame([0, 1], [1, 2]),
        "y": [1, 2, 3],
        "raises": AssertionError
    }
}


class TestDataSet(unittest.TestCase):
    def test_valid(self):
        for key, value in VALID.items():
            data = dataset.DataSet(value["X"], value["y"])
            self.assertEqual(list(data.X.ravel()), list(value["XX"].ravel()))
            self.assertEqual(list(data.labels.ravel()), list(value["yy"].ravel()))

    def test_invalid(self):
        for key, value in INVALID.items():
            self.assertRaises(value["raises"], dataset.DataSet, *(value["X"], value["y"]))



from setuptools import setup, find_packages


# see https://setuptools.readthedocs.io/en/latest/setuptools.html#basic-use

setup(name='pycvi',
      version='0.0.13',
      description='Easily compute clustering cliterions',
      url='',
      author='Yohan Foucade',
      author_email='foucade@lipn.fr',
      license='MIT',
      packages=find_packages(),
      install_requires=['numpy', 'pandas', 'scikit-learn'])
